/*
 * This file is based on code from the Dynamic Top Bar extension by Baudouin
 * Feildel (as at commit 603fa7eff3f06b6e9b136a7541b8476d28b5cea5).
 *
 * Licensing information:
 * Dynamic Top Bar Gnome Shell extension is distributed under the terms
 * of the GNU General Public License, version 3 or later.
 */

const GdkPixbuf = imports.gi.GdkPixbuf;
const Gio = imports.gi.Gio;
const Lang = imports.lang;
const Main = imports.ui.main;
const Mainloop = imports.mainloop;
const St = imports.gi.St;

const ExtensionUtils = imports.misc.extensionUtils;
const Me = ExtensionUtils.getCurrentExtension();
const Convenience = Me.imports.convenience;

const MAX_OPACITY = 255;

var zorinPanelDynamicTransparency = new Lang.Class({
    Name: 'ZorinPanel.DynamicTransparency',

    _init: function(settings) {
        this._zorinPanelSettings = settings;
    },

    enable: function() {
        // Making sure the extension activates after the User Theme has loaded
        Mainloop.idle_add(Lang.bind(this, function() {
            this.mainPanelTransparencyManager = new PanelTransparencyManager(
                [Main.panel.actor, Main.panel._leftCorner.actor, Main.panel._rightCorner.actor],
                this._zorinPanelSettings.get_int('panel-opacity')
            );
            new ShellManager(this.mainPanelTransparencyManager, this._zorinPanelSettings);

            return false;
        }));
    },

    enableWindowList: function() {
        this.windowListPanelTransparencyManager = new PanelTransparencyManager(
            [Main.windowList.actor],
            this._zorinPanelSettings.get_int('panel-opacity')
        );
        new ShellManager(this.windowListPanelTransparencyManager, this._zorinPanelSettings);
    },

    disable: function() {
        if (!this.mainPanelTransparencyManager)
        {
            this.mainPanelTransparencyManager = new PanelTransparencyManager(
                [Main.panel.actor, Main.panel._leftCorner.actor, Main.panel._rightCorner.actor], MAX_OPACITY);
        }
        this.mainPanelTransparencyManager.setSolid();
        this.mainPanelTransparencyManager = null;

        if (this.windowListPanelTransparencyManager) {
            this.disableWindowList();
        }
    },

    disableWindowList: function() {
        if (!this.windowListPanelTransparencyManager)
        {
            this.windowListPanelTransparencyManager = new PanelTransparencyManager(
                [Main.windowList.actor], MAX_OPACITY);
        }

        this.windowListPanelTransparencyManager.setSolid();
        this.windowListPanelTransparencyManager = null;
    }
});

/*
 * Panel transparency manager
 */
var PanelTransparencyManager = new Lang.Class({
    Name: 'PanelTransparencyManager',

    /**
     * Constructor
     * @param  {Object} panel The panel to manage
     */
    _init: function(actors, transparencyLevel) {
        this._actors = actors;
        this._isTransp = true;
        this._transparencyLevel = transparencyLevel;

        this._themeContextListener = St.ThemeContext.get_for_stage(global.stage).connect("changed", Lang.bind(this, this._userThemeChanged));

        this.updateColor();
    },

    _userThemeChanged: function() {
        this._resetStyle(this._actors);
        this.updateColor();

        if (this._isTransp)
            this.setTransparent();
        else
            this.setSolid();
    },

    updateColor: function() {
        let theme = this._actors[0].get_theme_node();
        let image_background = get_background_image_color(theme, this._actors[0]);
        let theme_background = theme.get_background_color();

        let background = (image_background !== null ? image_background : theme_background);

        this._color = {
            'r': background.red,
            'g': background.green,
            'b': background.blue,
            'a': this._transparencyLevel / MAX_OPACITY
        };
    },

    updateTransparencyLevel: function(newLevel) {
        let retransp = false;
        if(this._isTransp) {
            this.setSolid();
            retransp = true;
        }

        if(newLevel >= 0 && newLevel <= MAX_OPACITY)
            this._color.a = newLevel / MAX_OPACITY;

        if(retransp) this.setTransparent();
    },

    /**
     * Set the panel transparent
     */
    setTransparent: function(source) {
        // Add transparency
        this._isTransp = true;
        this._applyStyle();
    },

    /**
     * Set the panel solid
     */
    setSolid: function(source) {
        // Restore opacity
        this._isTransp = false;
        this._resetStyle(this._actors);
    },

    _setStyle: function(styles) {
        for (var i = 0; i < styles.length; i++) {
            if (typeof styles[i][0].set_style == 'function') {
                if (typeof styles[i][1] == 'string')
                    styles[i][0].set_style(styles[i][1]);
                else
                    styles[i][0].set_style('');
            }
        }
    },

    _resetStyle: function(styles) {
        for (var i = 0; i < styles.length; i++) {
            styles[i].set_style('');
        }
    },

    _applyStyle: function() {
        let toSet = null;

        let style = '';

        style = 'background-color: rgba('+this._color.r+','+this._color.g+','+this._color.b+','+this._color.a+');';
        style += 'border-color: rgba('+this._color.r+','+this._color.g+','+this._color.b+','+this._color.a+');';

        toSet = [[this._actors[0], style]];

        if (this._actors.length > 1) {
            let corner_style = '';
            corner_style = '-panel-corner-background-color: rgba('+this._color.r+','+this._color.g+','+this._color.b+','+this._color.a+');';

            toSet.push([this._actors[1], corner_style]);
            toSet.push([this._actors[2], corner_style]);
        }

        this._setStyle(toSet);
    },

    destroy: function() {
        St.ThemeContext.get_for_stage(global.stage).disconnect(this._themeContextListener);

        this._themeContextListener = null;
    }
});

/*
 * Window Manager
 * This class provide interface to work with windows
 */
var WindowManager = new Lang.Class({
    Name: 'WindowManager',

    /**
     * Constructor
     * @param  {Object} window    The window to manage
     * @param  {Object} workspace The workspace containing the window
     */
    _init: function(window, workspace) {
        this._metaWindow = window;
        this._workspace = workspace;

        this._notifyMinimizeId = this._metaWindow.connect('notify::minimized', Lang.bind(this, this._update));
        this._notifyMaximizeVId = this._metaWindow.connect('notify::maximized-vertically', Lang.bind(this, this._update));
        this._notifyMaximizeHId = this._metaWindow.connect('notify::maximized-horizontally', Lang.bind(this, this._update));
    },

    /**
     * Require the workspace to update it's state
     * This function is called on maximize, minize and restore events
     */
    _update: function() {
        this._workspace.updatePanelTransparency();
    },

    /**
     * Test if the window is maximized
     * @return {Boolean} true is window is maximized, false otherwise
     */
    isMaximized: function() {
        return this._metaWindow.maximized_vertically && !this._metaWindow.minimized;
    },

    /**
     * Get the monitor index which contain the window
     * @return {Number} The monitor index
     */
    getMonitorIndex: function() {
        return global.screen.get_monitor_index_for_rect(this._metaWindow.get_frame_rect());
    },

    /**
     * Test if the object manage metaWindow
     * @param  {Object} metaWindow The window to test
     * @return {Boolean}           true if metaWindow is managed by the object, false otherwise
     */
    equals: function(metaWindow) {
        return metaWindow == this._metaWindow;
    },

    /**
     * Destructor, removes event manager
     */
    destroy: function() {
        this._metaWindow.disconnect(this._notifyMinimizeId);
        this._metaWindow.disconnect(this._notifyMaximizeHId);
        this._metaWindow.disconnect(this._notifyMaximizeVId);
    }
});

/*
 * Workspace Manager
 */
var WorkspaceManager = new Lang.Class({
    Name: 'WorkspaceManager',

    _init: function(transparencyManager, metaWorkspace) {
        this._transparencyManager = transparencyManager;
        this._metaWorkspace = metaWorkspace;
        this._windowList = this.getWindowList();
        this._primaryMonitor = Main.layoutManager.primaryMonitor.index;

        this._notifyWindowAddedId = this._metaWorkspace.connect('window-added', Lang.bind(this, this._addWindow));
        this._notifyWindowRemovedId = this._metaWorkspace.connect('window-removed', Lang.bind(this, this._removeWindow));
    },

    getWindowList: function() {
        let tmpList = this._metaWorkspace.list_windows();
        let outList = [];

        for (let i = 0; i < tmpList.length; i++)
            outList.push(new WindowManager(tmpList[i], this));

        return outList;
    },

    isAnyWindowMaximized: function() {
        for (let i = 0; i < this._windowList.length; i++) {
            if (this._windowList[i].isMaximized() && this._windowList[i].getMonitorIndex() == this._primaryMonitor)
                return true;

            // Support DropDownTerminal https://github.com/zzrough/gs-extensions-drop-down-terminal
            if (this._windowList[i]._metaWindow.get_wm_class() == 'DropDownTerminalWindow')
                return true;
        }
        return false;
    },

    updatePanelTransparency: function(source) {
        if (this.isAnyWindowMaximized())
            this._transparencyManager.setSolid('WorkspaceManager.updatePanelTransparency');
        else
            this._transparencyManager.setTransparent('WorkspaceManager.updatePanelTransparency');
    },

    _haveWindow: function(metaWindow) {
        for (let i = 0; i < this._windowList.length; i++)
            if (this._windowList[i].equals(metaWindow))
                return true;
        return false;
    },

    _addWindow: function(metaWorkspace, metaWindow) {
        if (metaWindow.get_wm_class() == "Gnome-shell")
            return;

        if (!this._haveWindow(metaWindow))
            this._windowList.push(new WindowManager(metaWindow, this));

        this.updatePanelTransparency('WindowManager._addWindow, window:' + metaWindow.get_wm_class());
    },

    _removeWindow: function(metaWorkspace, metaWindow) {
        if (metaWindow.get_wm_class() == "Gnome-shell")
            return;

        for (let i = 0; i < this._windowList.length; i++)
            if (this._windowList[i].equals(metaWindow)) {
                this._windowList[i].destroy();
                this._windowList.splice(i, 1);
            }

        this.updatePanelTransparency('WindowManager._removeWindow, window:' + metaWindow.get_wm_class());
    },

    _onDestroy: function() {
        for (let i = 0; i < this._windowList.length; i++)
            this._windowList[i].destroy();

        this._metaWorkspace.disconnect(this._notifyWindowAddedId);
        this._metaWorkspace.disconnect(this._notifyWindowRemovedId);
    },

    destroy: function() {
        this._onDestroy();
        this._transparencyManager = null;
        this._metaWorkspace = null;
        this._windowList = null;
    }
});

var ShellManager = new Lang.Class({
    Name: 'ShellManager',

    _init: function(transparencyManager, settings) {
        this._settings = settings;
        this._transparencyManager = transparencyManager;
        this._currentWorkspace = new WorkspaceManager(transparencyManager, global.screen.get_active_workspace());
        this._currentWorkspace.updatePanelTransparency('ShellManager._init');

        this._notifySwitchId = global.window_manager.connect('switch-workspace', Lang.bind(this, this._switchWorkspace));
        this._notifyShowOverviewId = Main.overview.connect('showing', Lang.bind(this, this._showOverview));
        this._notifyHideOverviewId = Main.overview.connect('hiding', Lang.bind(this, this._hideOverview));
        this._overviewOpen = false;

        this._notifySettingsId = this._settings.connect('changed', Lang.bind(this, this._settingsUpdate));
    },

    _switchWorkspace: function(winManager, previousWkId, newWkId) {
        this._currentWorkspace.destroy();
        delete this._currentWorkspace;
        this._currentWorkspace = new WorkspaceManager(this._transparencyManager, global.screen.get_workspace_by_index(newWkId));
        if (this._overviewOpen)
            this._transparencyManager.setTransparent('ShellManager._switchWorkspace');
        else
            this._currentWorkspace.updatePanelTransparency('ShellManager._switchWorkspace');
    },

    _showOverview: function() {
        this._transparencyManager.setTransparent('ShellManager._showOverview');
        this._overviewOpen = true;
    },

    _hideOverview: function() {
        this._currentWorkspace.updatePanelTransparency('ShellManager._hideOverview');
        this._overviewOpen = false;
    },

    _onDestroy: function() {
        global.window_manager.disconnect(this._notifySwitchId);
        Main.overview.disconnect(this._notifyShowOverviewId);
        Main.overview.disconnect(this._notifyHideOverviewId);
        this._settings.disconnect(this._notifySettingsId);
    },

    _settingsUpdate: function() {
        this._transparencyManager.updateTransparencyLevel(this._settings.get_int('panel-opacity'));
        this._currentWorkspace.updatePanelTransparency('ShellManager._settingsUpdate');
    }
});

/**
 * Taken from Dynamic Panel Transparency.
 * src: https://github.com/rockon999/dynamic-panel-transparency/blob/master/dynamic-panel-transparency%40rockon999.github.io/theming.js
 */
function get_background_image_color(theme, actor) {
    let file = theme.get_background_image();

    if (!file) {
        // log('[Zorin Panel] No background image found in user theme.');
        let image = theme.get_border_image();

        if (!image) {
            // log('[Zorin Panel] No border image found in user theme.');
            return null;
        } else {
            file = image.get_file();
        }
    }

    try {
        let background = GdkPixbuf.Pixbuf.new_from_file(file.get_path());

        if (!background) {
            // log('[Zorin Panel] Provided background is null.');
            return null;
        }

        return average_color(background);
    } catch (error) {
        log('[Zorin Panel] Could not load the background and/or border image for your theme.');
        log(error);
        return null;
    }

}

/**
 * Taken from Plank. Used to calculate the average color of a theme's images.
 * src: http://bazaar.launchpad.net/~docky-core/plank/trunk/view/head:/lib/Drawing/DrawingService.vala
 *
 * @param {Object} source - A Gtk.Pixbuf
 */
function average_color(source, width, height) {
    let r, g, b, a, min, max;
    let delta;

    let rTotal = 0.0;
    let gTotal = 0.0;
    let bTotal = 0.0;

    let bTotal2 = 0.0;
    let gTotal2 = 0.0;
    let rTotal2 = 0.0;
    let aTotal2 = 0.0;

    let dataPtr = source.get_pixels();

    width = width ? width : source.get_width();
    height = width ? width : source.get_height();

    let length = width * height;

    let scoreTotal = 0.0;

    for (let y = 0; y < height; y++) {
        for (let x = 0; x < width; x++) {

            /* eslint-disable */

            let i = y * width * 4 + x * 4;

            r = dataPtr[i];
            g = dataPtr[i + 1];
            b = dataPtr[i + 2];
            a = dataPtr[i + 3];

            /* eslint-enable */

            // skip (nearly) invisible pixels
            if (a <= ALPHA_THRESHOLD) {
                length--;
                continue;
            }

            min = Math.min(r, Math.min(g, b));
            max = Math.max(r, Math.max(g, b));
            delta = max - min;

            // prefer colored pixels over shades of grey
            let score = SATURATION_WEIGHT * (delta === 0 ? 0.0 : delta / max);

            // weighted sums, revert pre-multiplied alpha value
            bTotal += score * b / a;
            gTotal += score * g / a;
            rTotal += score * r / a;
            scoreTotal += score;

            // not weighted sums
            bTotal2 += b;
            gTotal2 += g;
            rTotal2 += r;
            aTotal2 += a;
        }
    }

    // looks like a fully transparent image
    if (length <= 0) {
        return { red: 0, green: 0, blue: 0, alpha: 0 };
    }

    scoreTotal /= length;
    bTotal /= length;
    gTotal /= length;
    rTotal /= length;

    if (scoreTotal > 0.0) {
        bTotal /= scoreTotal;
        gTotal /= scoreTotal;
        rTotal /= scoreTotal;
    }

    bTotal2 /= length * 255;
    gTotal2 /= length * 255;
    rTotal2 /= length * 255;
    aTotal2 /= length * 255;

    // combine weighted and not weighted sum depending on the average "saturation"
    // if saturation isn't reasonable enough
    // s = 0.0 -> f = 0.0 ; s = WEIGHT_THRESHOLD -> f = 1.0
    if (scoreTotal <= WEIGHT_THRESHOLD) {
        let f = 1.0 / WEIGHT_THRESHOLD * scoreTotal;
        let rf = 1.0 - f;
        bTotal = bTotal * f + bTotal2 * rf;
        gTotal = gTotal * f + gTotal2 * rf;
        rTotal = rTotal * f + rTotal2 * rf;
    }

    // there shouldn't be values larger then 1.0
    let max_val = Math.max(rTotal, Math.max(gTotal, bTotal));
    if (max_val > 1.0) {
        bTotal /= max_val;
        gTotal /= max_val;
        rTotal /= max_val;
    }

    rTotal = Math.round(rTotal * 255);
    gTotal = Math.round(gTotal * 255);
    bTotal = Math.round(bTotal * 255);
    aTotal2 = Math.round(aTotal2 * 255);

    return { red: rTotal, green: gTotal, blue: bTotal, alpha: aTotal2 };
}
