/*
 * This code is taken from https://github.com/mlutfy/hidetopbar
 * (as at commit a2561b39d4a8e0b4438e94b18411cf8bf1889bb7)
 *
 * Licensing information:
 * Hide Top Bar extension is distributed under the terms of the
 * GNU General Public License, version 2 or later.
 */

const Lang = imports.lang;
const Mainloop = imports.mainloop;
const Meta = imports.gi.Meta;
const Shell = imports.gi.Shell;
const Clutter = imports.gi.Clutter;

const Main = imports.ui.main;
const Layout = imports.ui.layout;
const Tweener = imports.ui.tweener;
const Overview = imports.ui.overview;

const Me = imports.misc.extensionUtils.getCurrentExtension();
const Convenience = Me.imports.convenience;
const Intellihide = Me.imports.intellihide.intellihide;

const ShellActionMode = (Shell.ActionMode)?Shell.ActionMode:Shell.KeyBindingMode;

let ANIMATION_TIME = Overview.ANIMATION_TIME;
let PRESSURE_THRESHOLD = 100;
let PRESSURE_TIMEOUT = 1000;

var PanelVisibilityManager = new Lang.Class({
    Name: 'PanelVisibilityManager',

    _init: function(monitorIndex) {
        this._monitorIndex = monitorIndex;
        this.PanelBox = Main.windowList.actor;
        this._base_y = this.PanelBox.y;
        this._preventHide = false;
        this._staticBox = new Clutter.ActorBox();
        this._tweenActive = false;

        Main.layoutManager.removeChrome(this.PanelBox);
        Main.layoutManager.addChrome(this.PanelBox, {
            affectsStruts: false,
            trackFullscreen: true
        });

        this._initPressureBarrier();
        this._intellihide = new Intellihide.intellihide(this._monitorIndex);
        this._intellihide.enable();

        this._updateStaticBox();
        this._bindUIChanges();
    },

    hide: function(animationTime, trigger) {
        if(this._preventHide) return;

        let anchor_y = this.PanelBox.get_anchor_point()[1],
            delta_y = -this.PanelBox.height;
        if(anchor_y < 0) delta_y = -delta_y;
        let mouse = global.get_pointer(),
            mouse_is_over = (mouse[1] >= this._staticBox.y1 &&
                                 mouse[1] <= this._staticBox.y2 &&
                                 mouse[0] >= this._staticBox.x1 &&
                                 mouse[0] <= this._staticBox.x2);
        if(trigger == "mouse-left" && mouse_is_over) return;

        if(this._tweenActive) {
            Tweener.removeTweens(this.PanelBox, "y");
            this._tweenActive = false;
        }

        this._tweenActive = true;
        Tweener.addTween(this.PanelBox, {
            y: this._base_y + delta_y,
            time: animationTime,
            transition: 'easeOutQuad',
            onComplete: Lang.bind(this, function() {
                this._tweenActive = false;
                this.PanelBox.hide();
            })
        });
    },

    show: function(animationTime, trigger) {
        if(this._tweenActive) {
            Tweener.removeTweens(this.PanelBox, "y");
            this._tweenActive = false;
        }

        this.PanelBox.show();
        if(trigger == "destroy"
           || (
               trigger == "showing-overview"
               && global.get_pointer()[1] < this._panelHeight
              )
          ) {
            this.PanelBox.y = this._base_y;
        } else {
            this._tweenActive = true;
            Tweener.addTween(this.PanelBox, {
                y: this._base_y,
                time: animationTime,
                transition: 'easeOutQuad',
                onComplete: Lang.bind(this, function() {
                    this._tweenActive = false;
                    this._updateStaticBox();
                })
            });
        }
    },

    _handleMenus: function() {
        if(!Main.overview.visible) {
            let blocker = Main.windowList._menuManager.activeMenu;
            if(blocker == null) {
                this.hide(
                    ANIMATION_TIME,
                    "mouse-left"
                );
            } else {
                this._blockerMenu = blocker;
                this._menuEvent = this._blockerMenu.connect(
                    'open-state-changed',
                    Lang.bind(this, function(menu, open) {
                        if(!open && this._blockerMenu !== null) {
                            this._blockerMenu.disconnect(this._menuEvent);
                            this._menuEvent=null;
                            this._blockerMenu=null;
                            this._handleMenus();
                        }
                    })
                );
            }
        }
    },

    _disablePressureBarrier: function() {
        if(this._panelBarrier && this._panelPressure) {
            this._panelPressure.removeBarrier(this._panelBarrier);
            this._panelBarrier.destroy();
        }
    },

    _initPressureBarrier: function() {
        this._panelPressure = new Layout.PressureBarrier(
            PRESSURE_THRESHOLD,
            PRESSURE_TIMEOUT,
            ShellActionMode.NORMAL
        );
        this._panelPressure.connect(
            'trigger',
            Lang.bind(this, function(barrier) {
                if (Main.layoutManager.primaryMonitor.inFullscreen)
                    return;
                this.show(
                    ANIMATION_TIME,
                    "mouse-enter"
                );
            })
        );
        let anchor_y = this.PanelBox.get_anchor_point()[1],
            direction = Meta.BarrierDirection.POSITIVE_Y;
        if(anchor_y < 0) {
            anchor_y -= this.PanelBox.height;
            direction = Meta.BarrierDirection.NEGATIVE_Y;
        }
        this._panelBarrier = new Meta.Barrier({
            display: global.display,
            x1: this.PanelBox.x,
            x2: this.PanelBox.x + this.PanelBox.width,
            y1: this._base_y - anchor_y,
            y2: this._base_y - anchor_y,
            directions: direction
        });
        this._panelPressure.addBarrier(this._panelBarrier);
    },

    _updateStaticBox: function() {
        let anchor_y = this.PanelBox.get_anchor_point()[1];
        this._staticBox.init_rect(
            this.PanelBox.x, this.PanelBox.y-anchor_y, this.PanelBox.width, this.PanelBox.height
        );
        this._intellihide.updateTargetBox(this._staticBox);
    },

    _updateMouseSensitive: function() {
        this._disablePressureBarrier();
        this._initPressureBarrier();
    },

    _updatePreventHide: function() {
        this._preventHide = !this._intellihide.getOverlapStatus();
        let animTime = ANIMATION_TIME;
        if(this._preventHide)
            this.show(animTime, "intellihide");
        else if(!Main.overview.visible)
            this.hide(animTime, "intellihide");
    },

    _bindUIChanges: function () {
        this._signalsHandler = new Convenience.GlobalSignalsHandler();
        this._signalsHandler.add(
            [
                Main.overview,
                'showing',
                Lang.bind(this, function() {
                    this.show(
                        ANIMATION_TIME,
                        "showing-overview"
                    );
                })
            ],
            [
                Main.overview,
                'hiding',
                Lang.bind(this, function() {
                    this.hide(
                        ANIMATION_TIME,
                        "hiding-overview"
                    );
                })
            ],
            [
                this.PanelBox,
                'leave-event',
                Lang.bind(this, this._handleMenus)
            ],
            [
                this.PanelBox,
                'notify::anchor-y',
                Lang.bind(this, function () {
                    this._updateStaticBox();
                    this._updateMouseSensitive();
                })
            ],
            [
                global.screen,
                'monitors-changed',
                Lang.bind(this, function () {
                    this._base_y = this.PanelBox.y;
                    this._updateStaticBox();
                    this._updateMouseSensitive();
                })
            ],
            [
                this._intellihide,
                'status-changed',
                Lang.bind(this, this._updatePreventHide)
            ]
        );
    },

    destroy: function() {
        this._intellihide.destroy();
        this._signalsHandler.destroy();
        this._disablePressureBarrier();

        if (Main.windowList) {
            this.show(0, "destroy");
            Main.layoutManager.removeChrome(this.PanelBox);
            Main.layoutManager.addChrome(this.PanelBox, {
                affectsStruts: true,
                trackFullscreen: true
            });
        }

        this._intellihide = null;
        this._tweenActive = null;
        this._staticBox = null;
        this._preventHide = null;
        this._base_y = null;
        this.PanelBox = null;
        this._panelBarrier = null;
        this._panelPressure = null;
        this._signalsHandler = null;
    }
});
